/* -*- c++ -*-
 * TkResolver++ - Psychokinesis testing software
 *
 * Copyright (C) 2008 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "libtkcommon/intl.h"
#include "libtkcommon/debug.h"

#include "libtkquery/expression.h"
#include "libtkquery/identifier.h"
#include "libtkquery/parse_error.h"
#include "libtkdb/exception.h"

using namespace Tk::Query;

static const Tk::String obrkt("(");
static const Tk::String cbrkt(")");
static const Tk::String space(" ");
static const Tk::String in(" in ");

Boolean:: Boolean(bool b)
  : _val(b)
{
}

Query *
Boolean:: build_query(Tk::Db::Database *database, const Tk::String &domain) const
{
  //WDEBUG_OUT << L"building Boolean" << std::endl;
  Query *query( new Query(database) );
  
  Tk::Db::Data::Values params;
  params.push_back(_val);
  
  query->set_expr("?", params);
  
  return query;
}

Integer:: Integer(const char *text)
  : _val(atoi(text))
{
}

Query *
Integer:: build_query(Tk::Db::Database *database, const Tk::String &domain) const
{
  //WDEBUG_OUT << L"building Integer (" << _val << ")" << std::endl;
  Query *query( new Query(database) );

  Tk::Db::Data::Values params;
  params.push_back(_val);
  
  query->set_expr("?", params);
  
  return query;
}


Real:: Real(const char *text)
  : _val(strtod(text, NULL))
{
}

Query *
Real:: build_query(Tk::Db::Database *database, const Tk::String &domain) const
{
  //WDEBUG_OUT << L"building Real" << std::endl;
  Query *query( new Query(database) );

  Tk::Db::Data::Values params;
  params.push_back(_val);
  
  query->set_expr("?", params);
  
  return query;
}


String:: String(const Tk::String &str)
{
  // trim quote marks
  Tk::String::const_iterator c( str.begin() );
  for (++c; c != str.end();)
    {
      Tk::String::value_type chr( *c );
      
      if (++c != str.end())
	_val.push_back(chr);
    }
}
	
Query *
String:: build_query(Tk::Db::Database *database, const Tk::String &domain) const
{
  //WDEBUG_OUT << L"building String" << std::endl;
  Query *query( new Query(database) );

  Tk::Db::Data::Values params;
  params.push_back(_val);
  
  query->set_expr("?", params);
  
  return query;
}


Name:: Name(const Tk::String &name)
{
  // trim start and end marks
  Tk::String::const_iterator c( name.begin() );
  for (++c; c != name.end();)
    {
      Tk::String::value_type chr( *c );
      
      if (++c != name.end())
	_name.push_back(chr);
    }
}

const Tk::String &
Name:: get_name() const
{
  return _name;
}


Alpha:: Alpha (const char *text)
  : _text(text)
{
}

const Tk::String &
Alpha:: get_text() const
{
  return _text;
}


List::List(Value *value)
  : _values(1, value)
{
}

void
List:: append(Value *value)
{
  _values.push_back(value);
}

Query *
List:: build_query(Tk::Db::Database *database, const Tk::String &domain) const
{
  Query *query( new Query(database) );
  
  Tk::Db::Data::Values params;
  std::wostringstream expr_strm;
  
  for (ValueList::const_iterator iter( _values.begin() );
       iter != _values.end();
       ++iter)
    {
      QueryPtr val_query( (*iter)->build_query(database, domain) );
      
      query->append_tables(*val_query);
      params.append( val_query->get_expr_params() );
      
      if (iter != _values.begin())
	expr_strm << L", ";
      expr_strm << val_query->get_expr();
    }
  
  query->set_expr(expr_strm.str(), params);
  
  return query;
}


In:: In(Value *target, List *list)
  : _target(target),
    _list(list)
{
}


Query *
In:: build_query(Tk::Db::Database *database, const Tk::String &domain) const
{
  Query *query( new Query(database) );
  
  QueryPtr target_query( _target->build_query(database, domain) );
  query->append_tables(*target_query);
  Tk::Db::Data::Values params(target_query->get_expr_params() );
  
  QueryPtr list_query( _list->build_query(database, domain) );
  query->append_tables(*list_query);
  params.append( list_query->get_expr_params() );
  
  query->set_expr(target_query->get_expr() + in + obrkt + list_query->get_expr() + cbrkt, params);
  
  return query;
}


Op:: Op(const char *text)
  : _op(text)
{}


Binary:: Binary(Value *val1, Op *op, Value *val2)
  : _val1(val1),
    _op(op),
    _val2(val2)
{}

Binary:: ~Binary()
{
  delete _val1;
  delete _op;
  delete _val2;
}

Query *
Binary:: build_query(Tk::Db::Database *database, const Tk::String &domain) const
{
  //WDEBUG_OUT << L"building Binary" << std::endl;
  QueryPtr val1_query( _val1->build_query(database, domain) );
  QueryPtr val2_query( _val2->build_query(database, domain) );
  
  Query *query( new Query(database) );
  
  query->append_tables(*val1_query);
  query->append_tables(*val2_query);
  
  /* sandwich values and parameters */
  Tk::String expr( obrkt + val1_query->get_expr() + space + _op->get_op() + space + val2_query->get_expr() + cbrkt );
  Tk::Db::Data::Values expr_params( val1_query->get_expr_params() );
  expr_params.append(val2_query->get_expr_params());
  
  query->set_expr(expr, expr_params);
  
  return query;
}


Unary:: Unary (Op *op, Value *val)
  : _op(op),
    _val(val)
{}

Unary:: ~Unary()
{
  delete _op;
  delete _val;
}

Query *
Unary:: build_query(Tk::Db::Database *database, const Tk::String &domain) const
{
  //WDEBUG_OUT << L"building Unary" << std::endl;
  QueryPtr val_query( _val->build_query(database, domain) );
  
  Query *query( new Query(database) );
  query->append_tables(*val_query);

  Tk::Db::Data::Values params( val_query->get_expr_params() );
  params.push_back(false);
  
  query->set_expr(obrkt + val_query->get_expr() + Tk::String(" = ?") + cbrkt, params);
  
  return query;
}


Logical:: Logical (Expression *expr1, Op *lop, Expression *expr2)
  : _expr1(expr1),
    _lop(lop),
    _expr2(expr2)
{}

Logical:: ~Logical()
{
  delete _expr1;
  delete _lop;
  delete _expr2;
}

Query *
Logical:: build_query(Tk::Db::Database *database, const Tk::String &domain) const
{
  //WDEBUG_OUT << L"building Logical" << std::endl;
  Tk::String op;
  if (_lop->get_op() == "&")
    op = "and";
  else
    op = "or";
  
  QueryPtr expr1_query( _expr1->build_query(database, domain) );
  QueryPtr expr2_query( _expr2->build_query(database, domain) );
  
  Query *query( new Query(database) );
  
  query->append_tables(*expr1_query);
  query->append_tables(*expr2_query);
  
  /* sandwich values and parameters */
  Tk::String expr( expr1_query->get_expr() + space + op + space + expr2_query->get_expr() );
  Tk::Db::Data::Values expr_params( expr1_query->get_expr_params() );
  expr_params.append(expr2_query->get_expr_params());
  
  query->set_expr(expr, expr_params);
  
  return query;

  //  return _expr1->build_query(components) + space + op + space + _expr2->build_query(components);
}


Bracketed:: Bracketed (Expression *expr)
  : _expr(expr)
{}

Bracketed:: ~Bracketed()
{
  delete _expr;
}

Query *
Bracketed:: build_query(Tk::Db::Database *database, const Tk::String &domain) const
{
  //WDEBUG_OUT << L"building Bracketed" << std::endl;
  Query *query( _expr->build_query(database, domain) );
  query->set_expr( obrkt + query->get_expr() + cbrkt, query->get_expr_params() );
  return query;
}
