/*
 * TkResolver++ - Psychokinesis testing software
 *
 * Copyright (C) 2004,2009 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "libtkrandom/usbdev.h"

using namespace Tk::Random;

static const unsigned short TIMEOUT                 = 5000; /* milliseconds */

const unsigned short USBDev::BLOCK_SIZE(64);

USBDev:: USBDev(unsigned count,
		struct usb_dev_handle **handle_ptr,
		pthread_mutex_t *lock,
		int endpoint,
		bool check_short_read)
  : Retriever(count),
    _handle_ptr(handle_ptr),
    _lock(lock),
    _endpoint(endpoint),
    _check_short_read(check_short_read)
{
}

void
USBDev:: lock()
{
  WDEBUG_OUT << L"Locking USB dev" << std::endl;
  pthread_mutex_lock(_lock);
  WDEBUG_OUT << L"Locked USB dev" << std::endl;
}

void
USBDev:: unlock()
{
  pthread_mutex_unlock(_lock);
  WDEBUG_OUT << L"Unlocked USB dev" << std::endl;
}

void
USBDev:: condition_data(char *data, unsigned len, char **data_out, unsigned *len_out)
{
  *data_out = data;
  *len_out  = len;
}

void
USBDev:: retrieve(Collector &collector)
{
  Buffer *buf;
  char raw_buf[BLOCK_SIZE];
  int nread;
  char *rand_buf;
  unsigned rand_len;
  bool cont(true);
  
  lock();
  while (cont && (!limited() || _remaining > 0))
    {
      nread = usb_bulk_read(*_handle_ptr, _endpoint, raw_buf, BLOCK_SIZE, TIMEOUT);
      
      if (_check_short_read && nread < BLOCK_SIZE)
	throw std::runtime_error("short read from USB device");
      
      condition_data(raw_buf, nread, &rand_buf, &rand_len);
      
      buf = new Buffer;
      for (unsigned i = 0;
	   i < rand_len && (!limited() || _remaining > 0);
	   ++i, --_remaining)
	{
	  buf->push_back(rand_buf[i]);
	}
      
      cont = collector.collect(buf);
    }
  unlock();
}


void
USBDev:: open(int vendor_id, int product_id)
{
  lock();
  
  if (*_handle_ptr)
    {
      unlock();
      return;
    }
  
  usb_init();
  usb_find_busses();
  usb_find_devices();
  
  struct usb_bus    *bus;
  struct usb_device *dev;
  bool found(false);

  for (bus = usb_get_busses (); bus; bus = bus->next)
    {
      for (dev = bus->devices; dev; dev = dev->next)
	if (dev->descriptor.idVendor  == vendor_id &&
	    dev->descriptor.idProduct == product_id)
	  {
	    found = true;
	    break;
	  }
      
      if (found)
	break;
    }
  
  if (!found)
    {
      unlock();
      throw std::runtime_error("could not find USB device");
    }

  std::string usb_device(std::string(bus->dirname) + "/" + dev->filename);

  struct usb_dev_handle * handle( usb_open(dev) );
  if (!handle)
    {
      unlock();
      throw std::runtime_error(std::string("could not open USB device `") + usb_device + "'");
    }

  int err( usb_claim_interface(handle, 0) );
  if (err < 0)
    {
      unlock();
      throw UnixError(std::string("could not claim interface on device `")
		      + usb_device
		      + "' (return value: "
		      + static_cast<std::string>(Tk::String(err))
		      + ")",
		      -err);
    }
  
  *_handle_ptr = handle;
  unlock();
}

int
USBDev:: control_msg(int request_type, int request, int value, int index, char *bytes, int size, int timeout)
{
  lock();
  int ret( usb_control_msg(*_handle_ptr, request_type, request, value, index, bytes, size, timeout) );
  unlock();
  
  if (ret < 0)
    throw std::runtime_error("error sending USB control message");
  
  return ret;
}

int
USBDev:: bulk_read(int ep, char *bytes, int size, int timeout)
{
  lock();
  int ret( usb_bulk_read(*_handle_ptr, ep, bytes, size, timeout) );
  unlock();
  
  if (ret < 0)
    throw UnixError ("error reading from USB endpoint", -ret);
  
  return ret;
}


bool
USBDevType:: is_true_randomness()
{
  return true;
}

unsigned
USBDevType:: get_preferred_block_size()
{
  return USBDev::BLOCK_SIZE;
}
