/* -*- c++ -*-
 * TkResolver++ - Psychokinesis testing software
 *
 * Copyright (C) 2009 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#define __STDC_LIMIT_MACROS
#include <stdint.h>
#include <math.h>

#include "libtkrandom/rng.h"
#include "libtkresolver/colour.h"
#include "libtkresolver/types.h"

using namespace Tk::Resolver;

Colour:: Colour(double h, double s, double v)
  : _h(h), _s(s), _v(v)
{
  /*
  WDEBUG_OUT << L"Colour (h: "
	     << _h
	     << L"; s: "
	     << _s
	     << L"; v: "
	     << _v
	     << L")"
	     << std::endl;
  */
}

double Colour::get_h() const { return _h; }
double Colour::get_s() const { return _s; }
double Colour::get_v() const { return _v; }

void
Colour:: increment(const Colour &target, unsigned divisor)
{
  _h += (target._h - _h) / divisor;
  _s += (target._s - _s) / divisor;
  _v += (target._v - _v) / divisor;
}


Colour::Component
Colour::p() const
{
  return UINT8_MAX * _v * (1 - _s);
}

Colour::Component
Colour::q(double f) const
{
  return UINT8_MAX * _v * (1 - f * _s);
}

Colour::Component
Colour::t(double f) const
{
  return UINT8_MAX * _v * (1 - (1 - f) * _s);
}

void
Colour:: get_rgb(Component &r, Component &g, Component &b) const
{
  // as per http://en.wikipedia.org/wiki/HSL_color_space
  double h60(_h / 60.0);
  int h60f( floor(h60) );
  int h_i( h60f % 6 );
  
  double f( h60 - h60f );
  
  Component v( _v * UINT8_MAX );
  switch (h_i)
    {
    case 0: r = v;    g = t(f); b = p();  break;
    case 1: r = q(f); g = v;    b = p();  break;
    case 2: r = p();  g = v;    b = t(f); break;
    case 3: r = p();  g = q(f); b = v;    break;
    case 4: r = t(f); g = p();  b = v;    break;
    case 5: r = v;    g = p();  b = q(f); break;
    }
}

SDLColour
Colour:: map_rgb(SDL_PixelFormat *fmt) const
{
  Component r, g, b;
  get_rgb(r, g, b);
  return SDL_MapRGB(fmt, r, g, b);
}


ColourGenerator::ColourGenerator(unsigned screen_height)
  : _screen_height(screen_height),
    _start(0),
    _line_count(0),
    _current(generate()),
    _lines_left(0),
    _target(generate())
{
}

Colour
ColourGenerator:: generate()
{
  return Colour(Tk::Random::random<unsigned>(     0, 3600000) / 10000.0,
		Tk::Random::random<unsigned>(400000, 1000000) / 1000000.0,
		Tk::Random::random<unsigned>(400000, 1000000) / 1000000.0);
}

Colour
ColourGenerator:: get_colour()
{
  if (_start == 0)
    _start = SDL_GetTicks();
  
  if (_lines_left > 0)
    {
      _current.increment(_target, _lines_left);
      --_lines_left;
    }
  else
    {
      _current = _target;
      _target = generate();
      
      WDEBUG_OUT << L"Target colour: [h: "
		 << _target.get_h()
		 << L"; s: "
		 << _target.get_s()
		 << L"; v: "
		 << _target.get_v()
		 << L"]";
      
      Time now( SDL_GetTicks() );
      Time time( now - _start );
      
      double colour_distance( fabs(_target.get_h() - _current.get_h()) / 360.0 );
      if (time == 0)
	{
	  _lines_left = _screen_height * colour_distance;
	}
      else
	{
	  double lines_per_second( static_cast<double>(_line_count) / time * 1000.0 );
	  _lines_left = 10.0 * lines_per_second * colour_distance;
	  
	  WDEBUG_STRM << L"; lines per second: "
		      << lines_per_second;
	}
      
      WDEBUG_STRM << L"; lines: "
		  << _lines_left
		  << std::endl;
    }
  
  ++_line_count;
  
  return _current;
}
