/*
 * TkResolver++ - Psychokinesis testing software
 *
 * Copyright (C) 2004 Robert Ham <rah@bash.sh>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#define __STDC_LIMIT_MACROS
#include <stdint.h>
#include <SDL/SDL_gfxPrimitives.h>

#include <cmath>

#include "libtkcommon/exceptions.h"
#include "libtkcommon/debug.h"
#include "display.h"
#include "sdl.h"

using namespace Tk::Resolver;

static const double PHI( 1.61803398874989484820458683436563811772030917980576 );
static const double PI( 3.141592653589793238 );
  

Display:: Display(unsigned              height,
		  unsigned              width,
		  unsigned              trace_count,
		  const DirectionValue &arrow_direction,
		  unsigned              total_time,
		  double                horizon)
  : _trace_count(trace_count),
    _arrow_direction(arrow_direction),
    _total_time(total_time * MILLISECONDS_PER_SECOND),
    _display(0),
    _traces(0),
    _arrow(0),
    _horizon(0),
    _l(0),
    _start_time(0),
    _state(Started),
    _line_count(0),
    _prev_y(-1),
    _prev_traces(0)
{
  SDL_ShowCursor(SDL_DISABLE);
  
  _display = SDL_SetVideoMode(width, height, 8, SDL_ANYFORMAT|SDL_FULLSCREEN|SDL_HWSURFACE);
  //_display = SDL_SetVideoMode(width, height, 8, SDL_ANYFORMAT|SDL_HWSURFACE);
  if (!_display) 
    throw std::runtime_error(std::string("could not initialise display: ") + SDL_GetError());
  
  _foreground  = SDL_MapRGB(_display->format, UINT8_MAX, UINT8_MAX, UINT8_MAX);
  _background  = SDL_MapRGB(_display->format, 0, 0, 0);
  
  create_traces();
  if (arrow_direction.is_valid())
    create_arrow();
  
  _horizon = get_height() * horizon;
  _l = get_height() - _horizon;

  // wait for the monitor to switch
  //sleep (2);
  
};

Display::~Display()
{
  SDL_FreeSurface(_traces);
  if (_arrow)
    SDL_FreeSurface(_arrow);
  if (_prev_traces)
    delete _prev_traces;

  /*SDL_FreeSurface(_display);*/
  /* Seems to be the only way to do this */
  SDL::reset();
}

SDL_Surface *
Display:: create_surface(Uint32 flags, unsigned width, unsigned height, Uint32 fill)
{
  SDL_PixelFormat *fmt( _display->format );
  SDL_Surface *surface( SDL_CreateRGBSurface(SDL_SWSURFACE|flags, width, height, fmt->BitsPerPixel,
					     fmt->Rmask, fmt->Gmask, fmt->Bmask, fmt->Amask) );
  if (surface == 0)
    throw std::runtime_error(std::string("could not create surface: ") + SDL_GetError());
  
  SDL::lock(surface);
  SDL_FillRect(surface, NULL, fill);
  SDL::unlock(surface);
  
  return surface;
}

void
Display:: create_traces()
{
  _traces = create_surface(SDL_SRCALPHA, get_width(), get_height(), _background);
}


void
Display:: create_arrow()
{
  const Direction direction(_arrow_direction);
  
  unsigned height( get_height() * 0.10 );
  unsigned width( height * PHI );
  
  _arrow = create_surface(0, width, height, _background);
  
  unsigned tip(height / 2);
  unsigned body_top(height / 4);
  unsigned body_height(height / 2);
  unsigned chunk(width / 3);
  unsigned body_width(width - chunk);
  static const double barb_angle( (15.0 * PI) / 180.0 );
  unsigned head_width( chunk + (tip * tan(barb_angle)) );
  
  Sint16 x[3] = { /* tip  */ direction == Left ? 0 : width,
		  /* barb */ direction == Left ? head_width : (width - head_width),
		  /* body */ direction == Left ? chunk : (width - chunk) };
  filledTrigonRGBA(_arrow,
		   /* x1 */ x[0],
		   /* y1 */ tip,
		   /* x2 */ x[1],
		   /* y2 */ 0,
		   /* x3 */ x[2],
		   /* y3 */ tip,
		   128, 128, 128, SDL_ALPHA_OPAQUE);
 
  filledTrigonRGBA(_arrow,
  		   /* x1 */ x[0],
		   /* y1 */ tip,
  		   /* x2 */ x[1],
  		   /* y2 */ height,
  		   /* x3 */ x[2],
  		   /* y3 */ tip,
  		   128, 128, 128, SDL_ALPHA_OPAQUE);
  
  boxRGBA(_arrow,
  	  /* x1 */ direction == Left ? chunk : 0,
	  /* y1 */ body_top,
	  /* x2 */ direction == Left ? width : (width - chunk),
	  /* y2 */ body_top + body_height,
	  128, 128, 128, SDL_ALPHA_OPAQUE);
  
  SDL_SetColorKey(_arrow, SDL_SRCCOLORKEY|SDL_RLEACCEL, _background);
  
  _arrow_dest.y = height/2;
  _arrow_dest.w = width;
  _arrow_dest.h = height;
  if (direction == Left)
    _arrow_dest.x = height/2;
  else
    _arrow_dest.x = get_width() - width - height/2;
}

unsigned
Display:: get_height() const
{
  return _display->h;
};

unsigned
Display:: get_width() const
{
  return _display->w;
};

Time
Display:: get_time() const
{
  return SDL_GetTicks() - _start_time;
}

Time
Display:: stop() const
{
  Time t( get_time() );
  if (t > _total_time)
    {
      return t;
    }
  return 0;
}

Display::Coordinates
Display:: calc_start_x() const
{
  Coordinates start_line;
  unsigned trace_separation( get_width() / (_trace_count + 1) );
  for (OffsetIndex i = 0; i < _trace_count; ++i)
    start_line.push_back( trace_separation * (i + 1) );
  
  return start_line;
}

int
Display:: truncate_x(int x) const
{
  if (x < 0)
    x = 0;
  else if (x > get_width())
    x = get_width();

  return x;
}

void
Display:: get_x_limits(OffsetIndex index, Coordinate &min, Coordinate &max) const
{
  int x;
  
  if (_x_history.size() == 0)
    throw std::logic_error("cannot get x limits from empty history");

  min = max = _x_history.front()[index];

  for (LineHistory::const_iterator iter( ++(_x_history.begin()) );
       iter != _x_history.end();
       ++iter)
    {
      x = (*iter)[index];
      
      if (x > max)
	max = x;
      else if (x < min)
	min = x;
    }
}

bool
Display:: draw_arrow_p(const SDL_Rect *trace_dests)
{
  if (!_arrow)
    return false;
  
  if (_line_count == 0)
    return true;
  
  Coordinate min, max;
  if (_arrow_direction == Left)
    return _arrow_dest.x + _arrow_dest.w > trace_dests[0].x;
  else
    {
      unsigned last( _trace_count - 1 );
      return _arrow_dest.x < trace_dests[last].x + trace_dests[last].w;
    }
}  

void
Display:: clear_previous()
{
  if (!_prev_traces)
    return;
  
  SDL::lock(_display);
  for (unsigned i(0); i < _trace_count; ++i)
    SDL_FillRect(_display, _prev_traces + i, _background);
  SDL::unlock(_display);
}

SDL_Rect *
Display:: calc_trace_dests(Coordinate y)
{
  SDL_Rect *blits( new SDL_Rect[_trace_count] );
  
  for (unsigned i(0); i < _trace_count; ++i)
    {
      blits[i].y = 0;
      blits[i].h = y;
      
      // FIXME: check whether histories overlap on x axis
      Coordinate x_max, x_min;
      get_x_limits(i, x_min, x_max);
      
      blits[i].x = x_min;
      blits[i].w = x_max - x_min + 1;
    }
  
  return blits;
}

void
Display:: scroll_traces(const SDL_Rect *trace_dests, const Coordinates &prev_x, const Coordinates &this_x)
{
  for (unsigned i(0); i < _trace_count; ++i)
    {
      SDL_Rect src(trace_dests[i]);
      src.y += 1;
      //src.h -= 1;
      
      SDL_Rect dest(trace_dests[i]);
      
      SDL_SetColorKey(_traces, 0, 0);
      int err( SDL_BlitSurface(_traces, &src, _traces, &dest) );
      if (err)
	throw std::runtime_error(std::string("could not blit for scroll: ") + SDL_GetError());
	  
      
      if (this_x[i] != prev_x[i])
	{
	  SDL_Rect old;
	  old.y = trace_dests[i].h;
	  old.h = old.w = 1;
	  old.x = prev_x[i];
	  SDL::lock(_traces);
	  SDL_FillRect(_traces, &old, _background);
	  SDL::unlock(_traces);
	}
    }
}

void
Display:: draw_pixels(const Line &line, const Coordinates &x, Coordinate y)
{
  Uint32 sdl_colours[_trace_count];
  for (unsigned i(0); i < _trace_count; ++i)
    {
      if (line[i].colour)
	sdl_colours[i] = line[i].colour->map_rgb(_traces->format);
      else
	sdl_colours[i] = _foreground;
    }
     
  SDL_Rect pixel;
  pixel.h = pixel.w = 1;
  pixel.y = y;
  SDL::lock(_traces);
  for (unsigned i(0); i < _trace_count; ++i)
    {
      pixel.x = x[i];
      SDL_FillRect(_traces, &pixel, sdl_colours[i]);
    }
  SDL::unlock(_traces);
}

void
Display:: compose_display(const SDL_Rect *trace_dests, bool scrolling)
{
  bool draw_arrow( draw_arrow_p(trace_dests) );
  
  if (draw_arrow)
    SDL_BlitSurface(_arrow, NULL, _display, &_arrow_dest);
  
  SDL_SetColorKey(_traces, SDL_SRCCOLORKEY, _background);
  for (unsigned i(0); i < _trace_count; ++i)
    {
      SDL_Rect dest(trace_dests[i]);
      SDL_BlitSurface(_traces, &dest, _display, &dest);
      SDL_UpdateRect(_display, trace_dests[i].x, trace_dests[i].y, trace_dests[i].w, trace_dests[i].h);
    }
  
  if (draw_arrow)
    SDL_UpdateRect(_display, _arrow_dest.x, _arrow_dest.y, _arrow_dest.w, _arrow_dest.h);
}

void
Display:: draw_line(const Line &line)
{
  OffsetIndex i;
  
  Coordinates prev_x;
  if (_line_count > 0)
    prev_x = _x_history.back();
  else
    prev_x = calc_start_x();
  
  Coordinate y;
  Coordinates x;
  {
    Line::const_iterator line_iter( line.begin() );
    Coordinates::iterator prev_iter( prev_x.begin() );
    for (; line_iter != line.end(); ++line_iter, ++prev_iter)
      x.push_back( truncate_x(*prev_iter + line_iter->offset) );
  }
  _x_history.push_back(x);
  
  bool scrolling(_state == Scrolling || _state == Finished);
  switch (_state)
    {
    case Scrolling:
      y = _prev_y;
      break;
      
    case Started:
    case Finishing:
      y = _prev_y + 1;
      break;

    case Finished:
      y = get_height() - 1;
      break;
    };
  
  SDL_Rect *trace_dests( calc_trace_dests(y) );
  
  //WDEBUG_OUT << "Clearing previous" << std::endl;
  clear_previous();
  if (scrolling)
    {
      //WDEBUG_OUT << "Scrolling traces" << std::endl;
      scroll_traces(trace_dests, prev_x, x);
    }
  //WDEBUG_OUT << "Drawing pixels" << std::endl;
  draw_pixels(line, x, y);
  //WDEBUG_OUT << "Composing display" << std::endl;
  compose_display(trace_dests, scrolling);
  //WDEBUG_OUT << "Done drawing" << std::endl;
  
  
  if (scrolling)
    _x_history.erase(_x_history.begin());
  
  if (_prev_traces)
    delete _prev_traces;
  _prev_traces = trace_dests;
  
  _prev_y = y;
  ++_line_count;
  
  update_state();
}

void
Display:: start()
{
  _start_time = SDL_GetTicks();
}

void
Display:: update_state()
{
  switch (_state)
    {
    case Started:
      if (_prev_y == _horizon)
	_state = Scrolling;
      break;
    
    case Scrolling:
      if (start_final_leg_p())
	_state = Finishing;
      break;
      
    case Finishing:
      if (_prev_y >= get_height())
	_state = Finished;
      break;
    }
}

bool
Display:: start_final_leg_p() const
{
  Time time( get_time() );
  double s(time / _line_count);
  Time t( s * _l );
  
  return time >= (_total_time - t);
}

Display::BlockList
Display:: create_count_down_blocks(unsigned count)
{
  unsigned total_height = static_cast<unsigned>( get_height() * 5.0/7.0 );
  unsigned top          = (get_height() - total_height) / 2;
  unsigned height       = total_height / (count*2 - 1);
  unsigned width        = static_cast<unsigned>( height * PHI );
  unsigned left         = static_cast<unsigned>( get_width()/2.0 - width/2.0 );
  
  BlockList list;
  for (unsigned i = 0; i < count; ++i)
    {
      SDL_Rect block;
      block.x = left;
      block.y = top + height*2*i;
      block.h = height;
      block.w = width;

      list.push_back(block);
    }

  return list;
}

bool
Display:: delay(unsigned milliseconds)
{
  Time start( SDL_GetTicks() );
  while (SDL_GetTicks() - start < milliseconds)
    {
      if (SDL::check_escape(true))
	throw Escaped();
      SDL_Delay(10);
    }
}

void
Display:: count_down(unsigned count)
{
  BlockList blocks(create_count_down_blocks(count));

  // draw blocks
  SDL::lock(_display);
  for (BlockList::iterator iter(blocks.begin());
       iter != blocks.end();
       ++iter)
    {
      SDL_FillRect(_display, &(*iter), _foreground);
    }
  SDL::unlock(_display);
  
  SDL_UpdateRect(_display, blocks.front().x, blocks.front().y, blocks.front().w, blocks.back().y + blocks.back().h - blocks.front().y);

  // cover them up again
  for (BlockList::reverse_iterator iter(blocks.rbegin());
       iter != blocks.rend();
       ++iter)
    {
      delay(1000);

      SDL::lock(_display);
      SDL_FillRect(_display, &(*iter), _background);
      SDL::unlock(_display);
      
      SDL_UpdateRect(_display, iter->x, iter->y, iter->w, iter->h);
    }
  
  delay(1000);
}
